import cupy as cp  # type: ignore

from series_intro_recognizer.config import Config
from series_intro_recognizer.tp.tp import GpuFloat, GpuFloatArray, GpuInt


@cp.fuse()  # type: ignore
def _compute_offsets_and_indices(offsets_diff: int, length: int, rate: int) \
        -> tuple[GpuFloat, GpuFloat, GpuInt, GpuInt, GpuInt, GpuInt]:
    offset1_secs = cp.maximum(0.0, offsets_diff / rate)
    offset2_secs = cp.maximum(0.0, -offsets_diff / rate)

    start_idx_audio1 = cp.maximum(0, offsets_diff)
    end_idx_audio1 = start_idx_audio1 + length
    start_idx_audio2 = cp.maximum(0, -offsets_diff)
    end_idx_audio2 = start_idx_audio2 + length

    return (offset1_secs, offset2_secs,
            start_idx_audio1, end_idx_audio1,
            start_idx_audio2, end_idx_audio2)


def align_fragments(best_offset1: GpuFloat, best_offset2: GpuFloat,
                    audio1: GpuFloatArray, audio2: GpuFloatArray,
                    cfg: Config) -> tuple[GpuFloatArray, GpuFloatArray, GpuFloat, GpuFloat]:
    """
    Aligns two audio fragments based on the best offsets found by the correlator.
    Returns two audio fragments with the same duration, where the best_offsets are placed at the same point.
    For example, if there are two audios with duration 4s and 13s, and the best offsets are 1s and 3s,
    the resulting audios will have duration 4s.
    The 1st audio will be truncated from 0s to 4s, and the 2nd audio will be truncated from 3s to 7s.
    :param best_offset1: Offset of the common part of the audio fragments in the first audio fragment.
    :param best_offset2: Offset of the common part of the audio fragments in the second audio fragment.
    :param audio1:
    :param audio2:
    :param cfg: Configuration.
    :return: Tuple of truncated audio fragments and the offsets in seconds.
    """
    offsets_diff = best_offset1 - best_offset2
    length = cp.min(cp.array([audio1.size, audio2.size]))

    (offset1_secs, offset2_secs,
     start_idx_audio1, end_idx_audio1,
     start_idx_audio2, end_idx_audio2) = _compute_offsets_and_indices(offsets_diff, length, cfg.rate)

    truncated_audio1 = audio1[start_idx_audio1:end_idx_audio1]
    truncated_audio2 = audio2[start_idx_audio2:end_idx_audio2]

    return truncated_audio1, truncated_audio2, offset1_secs, offset2_secs



from typing import TypeVar

import cupy as cp  # type: ignore

T1 = TypeVar('T1')
T2 = TypeVar('T2')
T3 = TypeVar('T3')

type GpuInt = cp.ndarray[cp.int32]
type GpuFloat = cp.ndarray[cp.float32]
type GpuFloatArray = cp.ndarray[GpuFloat]
type GpuStack[T1, T2, T3] = cp.ndarray[T1, T2, T3]



import logging
from typing import Iterator, Any

import cupy as cp  # type: ignore
import numpy as np

from series_intro_recognizer.config import Config
from series_intro_recognizer.helpers.cached_iterator import iterate_with_cache
from series_intro_recognizer.services.best_offset_finder import find_best_offset
from series_intro_recognizer.services.correlator.correlator import calculate_correlation, CrossCorrelationResult
from series_intro_recognizer.services.interval_improver import improve_interval
from series_intro_recognizer.services.offsets_calculator import find_offsets
from series_intro_recognizer.tp.interval import Interval
from series_intro_recognizer.tp.tp import GpuFloatArray

logger = logging.getLogger(__name__)


def _load_to_gpu_and_normalize(audio: np.ndarray[Any, np.dtype[np.float64]]) -> GpuFloatArray:
    if audio.shape[0] == 0:
        raise ValueError('Empty audio passed.')

    gpu_audio = cp.asarray(audio, dtype=cp.float32)
    gpu_audio = gpu_audio - cp.mean(gpu_audio)
    gpu_audio = gpu_audio / cp.max(cp.abs(gpu_audio))

    return gpu_audio


def _save_corr_result(file1: int, file2: int, result: CrossCorrelationResult, cfg: Config) -> None:
    if not cfg.save_intermediate_results:
        return

    logger.info('Saving correlations for %s and %s...', file1, file2)
    with open(f'correlations/{file1}_{file2}_{result[0]:.3f}_{result[1]:.3f}.csv', 'w') as f:
        results = []
        for corr in result[2]:
            results.append(f'{corr[0]},{corr[1]}')
        f.write('\n'.join(results))


def _save_offsets_result(file1: int, file2: int, result: tuple[int, int], cfg: Config) -> None:
    if not cfg.save_intermediate_results:
        return

    logger.info('Saving offsets for %s and %s...', file1, file2)
    with open(f'offsets/{file1}.csv', 'a') as f:
        f.write(f'{file2},{result[0]},{result[1]}\n')
    with open(f'offsets/{file2}.csv', 'a') as f:
        f.write(f'{file1},{result[0]},{result[1]}\n')


def _find_offsets_for_episode(idx1: int, audio1: np.ndarray[Any, np.dtype[np.float64]],
                              idx2: int, audio2: np.ndarray[Any, np.dtype[np.float64]],
                              cfg: Config) -> tuple[Interval, Interval] | None:
    if audio1.shape[0] < cfg.min_segment_length_beats or audio2.shape[0] < cfg.min_segment_length_beats:
        logger.warning('One of the audios is shorter than %s secs: %s, %s. Skipping.',
                       cfg.min_segment_length_sec, audio1.shape[0], audio2.shape[0])
        return None

    # Step 1: Adjust the audios and calculate the correlation
    corr_result = calculate_correlation(audio1, audio2, cfg)
    if corr_result is None:
        return None
    _save_corr_result(idx1, idx2, corr_result, cfg)

    # Step 2: Find common part and its offsets
    corr_by_beats = corr_result[2][:, 1]
    offsets_result = find_offsets(corr_by_beats, cfg)
    if offsets_result is None:
        return None
    _save_offsets_result(idx1, idx2, offsets_result, cfg)

    # Step 3: Calculate the final offsets
    offset1_secs, offset2_secs, _ = corr_result
    begin1_secs = float(offset1_secs + offsets_result[0] * cfg.precision_secs)
    end1_secs = float(offset1_secs + offsets_result[1] * cfg.precision_secs)
    begin2_secs = float(offset2_secs + offsets_result[0] * cfg.precision_secs)
    end2_secs = float(offset2_secs + offsets_result[1] * cfg.precision_secs)

    # Step 4: Create intervals
    interval1 = Interval(begin1_secs, end1_secs)
    interval2 = Interval(begin2_secs, end2_secs)

    # Step 4: Improve the intervals
    audio1_duration = audio1.shape[0] / cfg.rate
    audio2_duration = audio2.shape[0] / cfg.rate
    improved_interval1 = improve_interval(interval1, audio1_duration, cfg)
    improved_interval2 = improve_interval(interval2, audio2_duration, cfg)

    logger.debug('Found offsets: %s, %s for %s and %s',
                 improved_interval1, improved_interval2, idx1, idx2)

    return improved_interval1, improved_interval2


def _find_offsets_for_episodes(audios: Iterator[np.ndarray[Any, np.dtype[np.float64]]],
                               cfg: Config) -> dict[int, list[Interval]]:
    pairs = iterate_with_cache(map(_load_to_gpu_and_normalize, audios), cfg.series_window)
    results: dict[int, list[Interval]] = {}
    for pair1, pair2 in pairs:
        idx1, audio1 = pair1
        idx2, audio2 = pair2
        logger.info(f'Processing %s and %s...', idx1, idx2)

        results.setdefault(idx1, [])
        results.setdefault(idx2, [])

        result = _find_offsets_for_episode(idx1, audio1, idx2, audio2, cfg)
        if result is None:
            continue

        results[idx1].append(result[0])
        results[idx2].append(result[1])

    cp.get_default_memory_pool().free_all_blocks()

    return results


def _find_most_likely_offsets(offsets_by_files: dict[int, list[Interval]], cfg: Config) -> dict[int, Interval]:
    """
    Returns the most likely offsets for each audio file.
    """
    true_offsets_by_files: dict[int, Interval] = {}
    for idx, offsets in offsets_by_files.items():
        true_offsets_by_files[idx] = find_best_offset(offsets, cfg)
        logger.debug('For %s: %.1f, %.1f (%.1fs)',
                     idx,
                     true_offsets_by_files[idx].start, true_offsets_by_files[idx].end,
                     true_offsets_by_files[idx].end - true_offsets_by_files[idx].start)

    return true_offsets_by_files


def recognise_from_audio_samples(audios: Iterator[np.ndarray[Any, np.dtype[np.float64]]],
                                 cfg: Config) -> list[Interval]:
    """
    Recognises series openings from a list of audio arrays.
    :param audios: list of audio arrays
    :param cfg: configuration
    :return: list of recognised intervals
    """
    offsets_by_files = _find_offsets_for_episodes(audios, cfg)
    true_offsets = _find_most_likely_offsets(offsets_by_files, cfg)
    logger.info('Results: %s', true_offsets)

    results = list(true_offsets.values())

    # Check if all indices are present in the correct order
    for i in range(len(results)):
        assert results[i] == true_offsets[i]

    return results



import logging
from typing import Annotated

import cupy as cp  # type: ignore

from series_intro_recognizer.config import Config
from series_intro_recognizer.services.correlator.async_correlator import correlation_with_async_moving_window
from series_intro_recognizer.services.correlator.fragments_normalizer import align_fragments
from series_intro_recognizer.services.correlator.sync_correlator import correlation_with_sync_moving_window
from series_intro_recognizer.tp.tp import GpuFloatArray, GpuStack, GpuFloat

CrossCorrelationResult = Annotated[
    tuple[GpuFloat, GpuFloat, GpuStack[GpuFloatArray, GpuFloatArray, None]],
    'CrossCorrelationResult']

logger = logging.getLogger(__name__)


def _get_offsets_of_best_match_beat(audio1: GpuFloatArray, audio2: GpuFloatArray, cfg: Config) \
        -> tuple[GpuFloat, GpuFloat]:
    offsets_by_windows = correlation_with_async_moving_window(audio1, audio2, cfg)
    best_match = offsets_by_windows[cp.argmax(offsets_by_windows[:, 2])]

    return best_match[0], best_match[1]


def calculate_correlation(audio1: GpuFloatArray, audio2: GpuFloatArray, cfg: Config) -> CrossCorrelationResult | None:
    """
    Aligns two audios and calculates correlation.
    :param audio1: audio1
    :param audio2: audio2
    :param cfg: Config
    :return: CrossCorrelationResult or None
    """
    best_offset1, best_offset2 = _get_offsets_of_best_match_beat(audio1, audio2, cfg)

    truncated_audio1, truncated_audio2, offset1_secs, offset2_secs = \
        align_fragments(best_offset1, best_offset2, audio1, audio2, cfg)
    if (truncated_audio1.shape[0] == 0
            or truncated_audio2.shape[0] == 0
            or truncated_audio1.shape[0] != truncated_audio2.shape[0]):
        # I believe this is not possible, but just in case
        logger.error('Truncated audios have different lengths: %s, %s. Skipping.',
                     truncated_audio1.shape[0], truncated_audio2.shape[0])
        return None

    corr_by_beats = correlation_with_sync_moving_window(truncated_audio1, truncated_audio2, cfg)

    return offset1_secs, offset2_secs, corr_by_beats



from typing import TypeVar, Iterator

T = TypeVar('T')


def iterate_with_cache(orig_iter: Iterator[T], n: int) -> Iterator[tuple[tuple[int, T], tuple[int, T]]]:
    """
    Iterates over an iterator, yielding pairs of N elements with a cache of N elements.

    For example, if N=2, the iterator will yield pairs of elements (0, 1), (0, 2), (1, 2), (1, 3), (2, 3), ...

    1st element will be stored in the cache from 1st to Nth iteration, then it will be unloaded, and the N+1st element
    will be stored in the cache.
    :param orig_iter: Original iterator
    :param n: Number of subsequent elements to cache
    :return: Iterator of pairs of elements
    """
    # Buffer to store elements from the iterator, allowing for caching up to N future elements
    buffer = []
    index = 0  # To keep track of the index of each item

    # Populate the initial buffer
    try:
        for _ in range(n + 1):
            buffer.append((index, next(orig_iter)))
            index += 1
    except StopIteration:
        # In case there are fewer than N+1 elements in the iterator
        pass

    # Main loop to yield pairs
    while len(buffer) > 1:  # Need at least two items to make a pair
        current_index, current_content = buffer[0]

        # Create pairs with the next N elements, if available
        for i in range(1, min(n + 1, len(buffer))):
            next_index, next_content = buffer[i]
            yield (current_index, current_content), (next_index, next_content)

        # Remove the processed element from the buffer
        buffer.pop(0)

        # Try to add a new element to the buffer if possible
        try:
            buffer.append((index, next(orig_iter)))
            index += 1
        except StopIteration:
            # No more elements to fetch from the original iterator
            pass



import logging

import cupy as cp  # type: ignore

from series_intro_recognizer.config import Config
from series_intro_recognizer.tp.tp import GpuFloatArray, GpuFloat

logger = logging.getLogger(__name__)


def _get_threshold(corr_values: GpuFloatArray) -> GpuFloat | None:
    max_limit = cp.mean(corr_values) + 2 * cp.std(corr_values)
    filtered = corr_values[corr_values < max_limit]

    if filtered.shape[0] == 0:
        logger.warning('Fragments are the same. Skipping. Try to increase the samples length.')
        return None

    return cp.max(filtered) / 2


def _longest_sequence_with_gaps(arr: GpuFloatArray, max_gap_length: int) -> tuple[int, int]:
    kernel = cp.ElementwiseKernel(
        in_params='raw bool arr, int32 max_gap_length',
        out_params='int32 max_start, int32 max_end',
        operation='''
            int n = arr.size();
            int current_start = -1;
            int current_end = -1;
            int longest_start = -1;
            int longest_end = -1;
            int gap_length = 0;

            for (int i = 0; i < n; i++) {
                if (arr[i]) {
                    if (current_start == -1) {
                        current_start = i;
                    }
                    current_end = i;
                    gap_length = 0;
                } else {
                    if (current_start != -1) {
                        gap_length++;
                        if (gap_length > max_gap_length) {
                            if ((longest_start == -1) || (current_end - current_start > longest_end - longest_start)) {
                                longest_start = current_start;
                                longest_end = current_end;
                            }
                            current_start = -1;
                            current_end = -1;
                            gap_length = 0;
                        }
                    }
                }
            }

            if ((current_start != -1) && (current_end - current_start > longest_end - longest_start)) {
                longest_start = current_start;
                longest_end = current_end;
            }

            max_start = longest_start;
            max_end = longest_end;
        ''',
        name='longest_sequence_with_gaps_gpu'
    )

    max_start = cp.zeros(1, dtype=cp.int32)
    max_end = cp.zeros(1, dtype=cp.int32)

    kernel(arr, max_gap_length, max_start, max_end)

    return int(max_start[0]), int(max_end[0])


def find_offsets(corr_values: GpuFloatArray, cfg: Config) -> tuple[int, int] | None:
    threshold = _get_threshold(corr_values)
    if threshold is None:
        return None

    bools = cp.asarray(corr_values > threshold)
    start, end = _longest_sequence_with_gaps(bools, cfg.offset_searcher_sequential_intervals)

    # Try to include the next element, because the end is exclusive
    # However, it would be incorrect if the end is at the last element,
    # so we need to check if it is the case.
    return start, min(end + 1, corr_values.size)



# -*- coding: utf-8 -*-
"""
series_intro_recognizer

Comparing episodes of a series to find the opening/endings of the series.

This library receives a list of episodes, extracts the audio features of each
episode and compares them to find the common part of the series.

To reduce the number of comparisons, the library compares 4 sequential episodes.
The number of episodes to be compared can be changed in configuration.

See README.md for more information.
"""

from series_intro_recognizer.tp.interval import Interval
from series_intro_recognizer.processors.audio_samples import recognise_from_audio_samples
from series_intro_recognizer.processors.audio_files import recognise_from_audio_files
